from . import stacks, i_stack, table
from .tools import concurrent_exec, get_exports
from .log import logger
from .common import *


def create(iregion):
    name = iregion.cfg.stack[0]
    stack = i_stack.ibox_stack(name, {}, region=iregion.name)
    iregion.cfg.exports = get_exports(obj=iregion)
    result = stack.replicate(ssm_map=iregion.ssm_map, iregion=iregion)
    if result:
        print(result)

    return result


def update(iregion):
    w_stacks = stacks.get(obj=iregion)
    iregion.cfg.stacks = list(w_stacks.keys())
    iregion.cfg.exports = get_exports(obj=iregion)
    result = concurrent_exec(
        "replicate",
        w_stacks,
        i_stack,
        region=iregion.name,
        **{"ssm_map": iregion.ssm_map, "iregion": iregion}
    )
    print(result)

    return result


def delete(iregion):
    w_stacks = stacks.get(obj=iregion)
    iregion.cfg.stacks = list(w_stacks.keys())
    result = concurrent_exec(
        "replicate",
        w_stacks,
        i_stack,
        region=iregion.name,
        **{"ssm_map": iregion.ssm_map, "iregion": iregion}
    )
    print(result)

    return result


def show(iregion):
    w_stacks = stacks.get(obj=iregion)
    s_table = table.get(list(w_stacks.values()))
    print(s_table)

    return s_table



import time
import concurrent.futures
from traceback import print_exc
from .aws import myboto3
from . import cfg
from .log import logger
from .common import *


def show_confirm():
    if cfg.parallel or cfg.answer_yes:
        return True

    print("\nEnter [y] to continue or any other key to exit: ", end="")
    sys.stdin = open(0)
    answer = sys.stdin.read(1)

    if not answer or answer.lower() != "y":
        return False
    else:
        return True


def _pause_or_stop():
    if cfg.pause == 0:
        if not show_confirm():
            return True
    elif cfg.pause and cfg.pause > 0:
        time.sleep(cfg.pause)


def concurrent_exec(command, stacks, smodule, region=None, **kwargs):
    n_failed = 0
    do_exit = False
    data = {}
    func = getattr(smodule, "exec_command")

    jobs = cfg.jobs if cfg.jobs else len(stacks)
    if jobs == 0:
        return

    cfg.parallel = False if not cfg.parallel and jobs == 1 else True

    with concurrent.futures.ThreadPoolExecutor(max_workers=jobs) as executor:
        future_to_stack = {}
        for s, v in stacks.items():
            ex_sub = executor.submit(func, s, v, command, region, **kwargs)

            future_to_stack[ex_sub] = s

            if not cfg.parallel and list(stacks)[-1] != s:
                concurrent.futures.wait({ex_sub: s})
                if _pause_or_stop():
                    do_exit = True
                    break

        for future in concurrent.futures.as_completed(future_to_stack):
            stack = future_to_stack[future]
            try:
                data[stack] = future.result()
            except IboxError as e:
                data[stack] = e.args[0]
                n_failed += 1
            except Exception as e:
                print(f"{stack} generated an exception: {e}")
                print_exc()
                raise IboxError(e)

    if n_failed == len(stacks):
        raise IboxError(f"All failed:\n{data}")

    if do_exit:
        print(data)
        exit(0)
    else:
        return data


def get_exports(obj=None):
    logger.info("Getting CloudFormation Exports")
    exports = {}

    if not obj:
        boto3 = myboto3()
        client = boto3.client("cloudformation")
    else:
        boto3 = getattr(obj, "boto3")
        client = boto3.client("cloudformation")

    paginator = client.get_paginator("list_exports")
    response_iterator = paginator.paginate()
    for e in response_iterator:
        for export in e["Exports"]:
            name = export["Name"]
            value = export["Value"]
            exports[name] = value
        # if all(key in exports for key in ['BucketAppRepository']):
        #    return exports

    return exports


def stack_resource_to_dict(stack):
    out = {}
    for n in dir(stack):
        if not n.startswith("__"):
            prop = ""
            words = n.split("_")
            for w in words:
                prop += w.capitalize()
            out[prop] = getattr(stack, n)

    return out


def smodule_to_class(smodule):
    class obj(object):
        pass

    cls = obj()
    for n in dir(smodule):
        if not n.startswith("_"):
            value = getattr(smodule, n)
            setattr(cls, n, value)

    return cls



from .common import *


def get_action_tags(istack, stack_tags):
    # unchanged tags
    tags_default = {}

    # changed tags - same value as corresponding stack param
    tags_changed = {}
    final_tags = []

    for tag in stack_tags:
        key = tag["Key"]
        current_value = tag["Value"]

        # Skip LastUpdate and EnvApp1Version Tag
        if key in ["LastUpdate", "EnvApp1Version"]:
            continue

        # check if key exist as cfg param/attr too
        try:
            cfg_value = getattr(istack.cfg, key)
            in_cfg = True if cfg_value is not None else None
        except Exception:
            in_cfg = None

        # current value differ from cmd arg
        if in_cfg and current_value != cfg_value:
            value = cfg_value

            # tags value cannot be empty
            if len(value) == 0:
                value = "empty"

            tags_changed[key] = "%s => %s" % (current_value, value)

        # keep current tag value
        else:
            value = current_value

            # tags value cannot be empty
            if len(value) == 0:
                value = "empty"

            tags_default[key] = value

        final_tags.append({"Key": key, "Value": value})

    # Add LastUpdate Tag with current time
    # Currently disabled:
    # Some resource, like CloudFormation Distribution, take time to be updated.
    # Does it make sense to have a tag with LastUpdateTime even if resource properties are not changed at all ?
    # If a resource is created by CloudFormation i can simply look at Stack LastUpdateTime
    # to have the same information derived by tagging it (i know that with tagging is simpler to do this).
    # final_tags.append({"Key": "LastUpdate", "Value": str(datetime.now())})

    if len(tags_default) > 0:
        istack.mylog(
            "DEFAULT - STACK TAGS\n%s\n" % pformat(tags_default, width=1000000)
        )
    if len(tags_changed) > 0:
        istack.mylog(
            "CHANGED - STACK TAGS\n%s\n" % pformat(tags_changed, width=1000000)
        )

    return final_tags



# parser default cfg
stack = role = type = topics = stack_args = []
parallel = region = jobs = pause = version = template = None
nowait = compact = dryrun = answer_yes = no_stacks = all_stacks = None
debug = False
max_retry_ecs_service_running_count = 0
timedelta = 300
dashboard = "OnChange"
statistic = "Average"
statisticresponse = "p95"
silent = True
vertical = False
profile = False
output = "text"
disable_rollback = False
#

OUT_WIDTH = 1000000

SLACK_CHANNEL = "_cf_deploy"

MAX_SINGLE_STACKS = 5

ACTION_WAITER_SLEEP_TIME = 3

STACK_BASE_DATA = [
    "StackName",
    "Description",
    "StackStatus",
    "CreationTime",
    "LastUpdatedTime",
]

RESOURCES_MAP = {
    "AutoScalingGroup": "AutoScalingGroupName",
    "AutoScalingGroupSpot": "AutoScalingGroupSpotName",
    # 'TargetGroup': 'TargetGroup',
    # 'TargetGroupExternal': 'TargetGroupExternal',
    # 'TargetGroupInternal': 'TargetGroupInternal',
    "Service": "ServiceName",
    "ServiceSpot": "ServiceSpotName",
    "ServiceExternal": "ServiceName",
    "ServiceInternal": "ServiceName",
    "LoadBalancerClassicExternal": "LoadBalancerNameExternal",
    "LoadBalancerClassicInternal": "LoadBalancerNameInternal",
    "LoadBalancerApplicationExternal": "LoadBalancerExternal",
    "LoadBalancerApplicationInternal": "LoadBalancerInternal",
    "Cluster": "ClusterName",
    "ScalableTarget": "ClusterName",
    "ListenerHttpsExternalRules1": "LoadBalancerExternal",
    "ListenerHttpsExternalRules2": "LoadBalancerExternal",
    "ListenerHttpExternalRules1": "LoadBalancerExternal",
    "ListenerHttpInternalRules1": "LoadBalancerInternal",
    "AlarmCPUHigh": None,
    "AlarmCPULow": None,
    "AWS::ElasticLoadBalancingV2::TargetGroup": None,
}
SCALING_POLICY_TRACKINGS_NAMES = {
    "ScalingPolicyTrackings1": None,
    "ScalingPolicyTrackingsApp": None,
    "ScalingPolicyTrackingsASCpu": "ScalingPolicyTrackings1",
    "ScalingPolicyTrackingsASCustom": "ScalingPolicyTrackings1",
    "ScalingPolicyTrackingsAPPCpu": "ScalingPolicyTrackings1",
    "ScalingPolicyTrackingsAPPCustom": "ScalingPolicyTrackings1",
    "AutoScalingScalingPolicyCpu": "ScalingPolicyTrackings1",
    "AutoScalingScalingPolicyCustom": "ScalingPolicyTrackings1",
    "ApplicationAutoScalingScalingPolicyCpu": "ScalingPolicyTrackingsApp",
    "ApplicationAutoScalingScalingPolicyCustom": "ScalingPolicyTrackingsApp",
    # Disabled until they are properly managed in dashboard.py [get_policy].
    # Ideal should be to create two horizontal annotations
    # one for the lower bound and one for the upper one
    # but as AutoScalingScalingPolicy have been replaced with Tracking ones
    # i do not know if it make sense to put much effort in this.
    # 'AutoScalingScalingPolicyDown': 'ScalingPolicyTrackings1',
    # 'AutoScalingScalingPolicyUp': 'ScalingPolicyTrackings1',
    # 'ApplicationAutoScalingScalingPolicyDown': 'ScalingPolicyTrackings1',
    # 'ApplicationAutoScalingScalingPolicyUp': 'ScalingPolicyTrackings1',
}
RESOURCES_MAP.update(SCALING_POLICY_TRACKINGS_NAMES)

RESOURCES_MAP_R53 = {
    "RecordSetExternal": None,
    "RecordSetInternal": None,
    "RecordSetExternalRO": None,
    "RecordSetInternalRO": None,
    "RecordSetCloudFront": None,
    "AWS::ServiceDiscovery::Service": None,
}

STACK_COMPLETE_STATUS = [
    "UPDATE_COMPLETE",
    "CREATE_COMPLETE",
    "ROLLBACK_COMPLETE",
    "UPDATE_ROLLBACK_COMPLETE",
    "UPDATE_ROLLBACK_FAILED",
    "DELETE_COMPLETE",
    "DELETE_FAILED",
]

CHANGESET_COMPLETE_STATUS = [
    "CREATE_COMPLETE",
    "UPDATE_ROLLBACK_FAILED",
    "FAILED",
]

SHOW_TABLE_FIELDS = [
    "EnvStackVersion",
    "EnvRole",
    "StackName",
    "StackType",
    "UpdateMode",
    "EnvApp1Version",
    "StackStatus",
    "LastUpdatedTime",
]

STACKSET_INSTANCES_SHOW_TABLE_FIELDS = [
    "Region",
    "Account",
    "StackId",
    "Status",
    "StatusReason",
    "StackInstanceStatus",
]

SSM_BASE_PATH = "/ibox"



import argparse
from . import cfg
from .commands import (
    create,
    update,
    delete,
    cancel_update,
    continue_update,
    rollback,
    info,
    parameters,
    resolve,
    show_table,
    log,
    dash,
    ssm_setup,
    ssm_put,
    ssm_show,
    r53,
    replicate,
    stackset,
)


def get_create_parser(subparser, parents=[]):
    parser = subparser.add_parser("create", parents=parents, help="Create Stack")
    parser.add_argument("--Env", help="Environment to use", type=str, required=True)
    parser.add_argument("--EnvRole", help="Stack Role", type=str, required=True)
    parser.add_argument("--EnvApp1Version", help="App Version", type=str, default="")

    return parser


def get_update_parser(subparser, parents=[]):
    parser = subparser.add_parser("update", parents=parents, help="Update Stack")
    parser.add_argument(
        "-P",
        "--policy",
        help="Policy during Stack Update",
        type=str,
        choices=[
            "*",
            "Modify",
            "Delete",
            "Replace",
            "Modify,Delete",
            "Modify,Replace",
            "Delete,Replace",
        ],
    )
    parser.add_argument(
        "-n", "--nochangeset", help="No ChangeSet", required=False, action="store_true"
    )
    parser.add_argument("--dryrun", help="Show changeset and exit", action="store_true")
    parser.add_argument(
        "-T", "--showtags", help="Show tags changes in changeset", action="store_true"
    )
    parser.add_argument(
        "-D",
        "--dashboard",
        help="Update CloudWatch DashBoard",
        choices=["Always", "OnChange", "Generic", "None"],
        default=cfg.dashboard,
    )
    parser.add_argument(
        "--nodetails",
        help="Do not show extra details in changeset",
        action="store_true",
    )

    return parser


def set_dash_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        "dash", parents=parents, help="Create DashBoard for stacks"
    )
    parser.set_defaults(func=dash)

    parser.add_argument(
        "--statistic",
        help="Statistic to use for metrics",
        choices=["Average", "Maximum", "Minimum"],
        default="Average",
    )
    parser.add_argument(
        "--statisticresponse",
        help="Statistic to use for response time metrics",
        choices=["Average", "p99", "p95", "p90", "p50", "p10", "Maximum", "Minimum"],
        default="p95",
    )
    parser.add_argument("--debug", help="Show json Dash", action="store_true")
    parser.add_argument("--silent", help="Silent mode", action="store_true")
    parser.add_argument(
        "--vertical",
        help="Add vertical annotation at creation time, "
        "and optionally specify fill mode",
        nargs="?",
        choices=["before", "after"],
        const=True,
        default=False,
    )


def get_show_parser(subparser, parents=[]):
    parser = subparser.add_parser("show", parents=parents, help="Show Stacks table")

    parser.add_argument(
        "-F", "--fields", nargs="+", type=str, default=cfg.SHOW_TABLE_FIELDS
    )
    parser.add_argument(
        "-O", "--output", type=str, default="text", choices=["text", "html", "bare"]
    )

    return parser


def set_ssm_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        "ssm", parents=[], help="SSM Parameters override for Stack Replicator"
    )

    ssm_parser = parser.add_subparsers(
        title="SSM Command", required=True, dest="command_ssm"
    )
    # setup
    setup_parser = ssm_parser.add_parser("setup", help="Setup Regions", parents=parents)
    setup_parser.set_defaults(func=ssm_setup, no_stacks=True)

    setup_parser.add_argument(
        "-R",
        "--regions",
        help="Regions",
        type=str,
        required=True,
        default=[],
        nargs="+",
    )
    # put
    put_parser = ssm_parser.add_parser("put", help="Put Parameters", parents=parents)
    put_parser.set_defaults(func=ssm_put)

    put_parser.add_argument(
        "-R", "--regions", help="Regions", type=str, default=[], nargs="+"
    )
    # show
    show_parser = ssm_parser.add_parser(
        "show", help="Show Regions Distribution", parents=parents
    )
    show_parser.set_defaults(func=ssm_show, all_stacks=True)


def set_stackset_parser(subparser, parents=[]):
    parser = subparser.add_parser("stackset", parents=[], help="StackSet operations")
    parser.set_defaults(func=stackset)

    stackset_parser = parser.add_subparsers(
        title="Stackset Commands", required=True, dest="command_stackset"
    )

    # update parser
    parser_update = stackset_parser.add_parser(
        "update",
        parents=parents
        + [
            get_stackset_single_parser(),
            get_template_parser(required=False),
        ],
    )

    # info parser
    parser_info = stackset_parser.add_parser(
        "info",
        parents=[
            get_stackset_single_parser(),
            get_template_parser(required=False),
        ],
    )

    # parameters parser
    parser_parameters = stackset_parser.add_parser(
        "parameters",
        parents=[
            get_stackset_single_parser(),
            get_template_parser(required=False),
        ],
    )

    # show parser
    parser_show = get_show_parser(
        stackset_parser,
        [
            get_stackset_single_parser(),
        ],
    )

    # instances parser
    parser_instances = stackset_parser.add_parser(
        "instances",
        parents=[
            get_stackset_single_parser(),
            get_template_parser(required=False),
        ],
    )


def set_replicate_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        "replicate", parents=[], help="Replicate in Regions configured by SSM"
    )
    parser.set_defaults(func=replicate)

    replicate_parser = parser.add_subparsers(
        title="Replicate Command", required=True, dest="command_replicate"
    )

    regions_parser = argparse.ArgumentParser(add_help=False)
    regions_parser.add_argument(
        "-R", "--regions", help="Regions", type=str, default=[], nargs="+"
    )
    regions_parser.add_argument(
        "--no_replicate_current",
        help="No replication in current Region",
        action="store_true",
    )

    # replicate create
    parser_create = get_create_parser(
        replicate_parser,
        parents
        + [
            get_template_parser(),
            get_stack_single_parser(),
            get_create_update_parser(),
            regions_parser,
        ],
    )

    # replicate update
    parser_update = get_update_parser(
        replicate_parser,
        parents
        + [
            get_template_parser(required=False),
            get_stack_selection_parser(),
            get_create_update_parser(),
            regions_parser,
        ],
    )

    # replicate delete
    parser_delete = replicate_parser.add_parser(
        "delete",
        parents=parents + [get_stack_single_parser(), regions_parser],
        help="Delete Stack (WARNING)",
    )

    # replicate show
    parser_show = get_show_parser(
        replicate_parser, [get_stack_selection_parser(), regions_parser]
    )


def set_r53_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        "r53",
        parents=parents,
        help="Create RecordSet Aliases looking at stack R53 resources",
    )
    parser.set_defaults(func=r53)

    parser.add_argument("--dryrun", help="Show changes and exit", action="store_true")
    parser.add_argument(
        "--noorigin", help="Do not create Origin Record", action="store_true"
    )
    parser.add_argument(
        "--suffix",
        help='Suffix to add to RecordSet Name, prepended with "-"',
        default="",
        type=str,
    )


def get_template_parser(required=True):
    parser = argparse.ArgumentParser(add_help=False)

    group = parser.add_mutually_exclusive_group(required=required)
    group.add_argument("--template", help="Template Location", type=str)
    group.add_argument("-v", "--version", help="Stack Env Version", type=str)

    return parser


def get_stack_selection_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        "-s",
        "--stack",
        nargs="+",
        help="Stack Names space separated",
        type=str,
        default=[],
    )
    parser.add_argument(
        "-r",
        "--role",
        nargs="+",
        help="Stack Roles space separated",
        type=str,
        default=[],
    )
    parser.add_argument(
        "-t",
        "--type",
        nargs="+",
        help="Stack Types space separated - use ALL for any type",
        type=str,
        default=[],
    )

    return parser


def get_stack_single_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        "-s",
        "--stack",
        nargs=1,
        help="Stack Names space separated",
        required=True,
        type=str,
        default=[],
    )

    return parser


def get_stackset_single_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        "-s",
        "--stackset",
        nargs=1,
        dest="stack",
        help="StackSet name",
        required=True,
        type=str,
        default=[],
    )

    return parser


def get_create_update_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        "--topics",
        nargs="+",
        help="SNS Topics Arn for notification",
        type=str,
        default=[],
    )
    parser.add_argument(
        "-M",
        "--max_retry_ecs_service_running_count",
        help="Trigger Stack Cancel Update (RollBack) if Service deployment's failedTasks is equals or greater than max_retry",
        type=int,
        default=0,
    )
    parser.add_argument(
        "--disable_rollback",
        help="Preserve the state of previously provisioned "
        "resources when an operation fails",
        action="store_true",
    )

    return parser


# parse main argumets
def get_parser():
    parser = argparse.ArgumentParser(
        description="Stacks Operations",
        epilog="Note: options for Stack Params must be put at the end!",
    )

    # common parser
    parser.add_argument("--region", help="Region", type=str)
    parser.add_argument("--profile", help="AWS Profile", type=str)
    parser.add_argument(
        "--compact", help="Display Stacks-Output in compact form", action="store_true"
    )
    parser.add_argument(
        "-j",
        "--jobs",
        help="Max Concurrent jobs - default to number of stacks",
        type=int,
    )
    parser.add_argument(
        "--pause",
        help="Pause for seconds between jobs - "
        "0 for interactive - valid only for jobs=1",
        type=int,
    )

    # action parser
    action_parser = argparse.ArgumentParser(add_help=False)

    action_parser.add_argument(
        "-y",
        "--answer_yes",
        help="Answer YES (No Confirm)",
        required=False,
        action="store_true",
    )
    action_parser.add_argument(
        "-w",
        "--nowait",
        help="Do not Wait for action to end",
        required=False,
        action="store_true",
    )
    action_parser.add_argument(
        "-c",
        "--slack_channel",
        help=f"Slack Channel [{cfg.SLACK_CHANNEL}]",
        nargs="?",
        const=cfg.SLACK_CHANNEL,
        default=False,
    )

    # template parser
    template_parser_create = get_template_parser()
    template_parser_update = get_template_parser(required=False)

    # stack selection parser
    stack_selection_parser = get_stack_selection_parser()

    # stack single parser
    stack_single_parser = get_stack_single_parser()

    # create update parser
    create_update_parser = get_create_update_parser()

    # command subparser
    command_subparser = parser.add_subparsers(
        help="Desired Command", required=True, dest="command"
    )

    # create parser
    parser_create = get_create_parser(
        command_subparser,
        [
            action_parser,
            template_parser_create,
            stack_single_parser,
            create_update_parser,
        ],
    )
    parser_create.set_defaults(func=create)

    # update parser
    parser_update = get_update_parser(
        command_subparser,
        [
            action_parser,
            template_parser_update,
            stack_selection_parser,
            create_update_parser,
        ],
    )
    parser_update.set_defaults(func=update)

    # delete parser
    parser_delete = command_subparser.add_parser(
        "delete",
        parents=[action_parser, stack_single_parser],
        help="Delete Stack (WARNING)",
    )
    parser_delete.set_defaults(func=delete)

    # cancel_update parser
    parser_cancel = command_subparser.add_parser(
        "cancel",
        parents=[action_parser, stack_selection_parser],
        help="Cancel Update Stack",
    )
    parser_cancel.set_defaults(func=cancel_update)

    # continue_update parser
    parser_continue = command_subparser.add_parser(
        "continue",
        parents=[action_parser, stack_selection_parser],
        help="Continue Update RollBack",
    )
    parser_continue.set_defaults(func=continue_update)
    parser_continue.add_argument(
        "--resources_to_skip", help="Resource to Skip", default=[], nargs="+"
    )

    # roll back parser
    parser_rollback = command_subparser.add_parser(
        "rollback",
        parents=[action_parser, stack_selection_parser],
        help="RollBack to the last known stable state",
    )
    parser_rollback.set_defaults(func=rollback)

    # info parser
    parser_info = command_subparser.add_parser(
        "info", parents=[stack_selection_parser], help="Show Stack Info"
    )
    parser_info.set_defaults(func=info)

    # parameters parser
    parser_parameters = command_subparser.add_parser(
        "parameters",
        parents=[template_parser_update, stack_selection_parser],
        help="Show Available Stack Parameters",
    )
    parser_parameters.set_defaults(func=parameters)

    # resolve parser
    parser_resolve = command_subparser.add_parser(
        "resolve",
        parents=[template_parser_update, stack_selection_parser],
        help="Resolve Stack template - output in yaml short format",
    )
    parser_resolve.set_defaults(func=resolve)

    # log parser
    parser_log = command_subparser.add_parser(
        "log", parents=[stack_single_parser], help="Show Stack Log"
    )
    parser_log.set_defaults(func=log)
    parser_log.add_argument(
        "-d",
        "--timedelta",
        help="How many seconds go back in time from stack last event - "
        "use 0 for realtime - if < 30 assume days",
        default=cfg.timedelta,
    )

    # dashboard parser
    set_dash_parser(
        command_subparser,
        [
            stack_selection_parser,
        ],
    )

    # show parser
    parser_show = get_show_parser(
        command_subparser,
        [
            stack_selection_parser,
        ],
    )
    parser_show.set_defaults(func=show_table, all_stacks=True)

    # stackset parser
    set_stackset_parser(
        command_subparser,
        [
            action_parser,
        ],
    )

    # ssm parser
    set_ssm_parser(
        command_subparser,
        [
            stack_selection_parser,
        ],
    )

    # replicate parser
    set_replicate_parser(
        command_subparser,
        [
            action_parser,
        ],
    )

    # r53 parser
    set_r53_parser(
        command_subparser,
        [
            stack_selection_parser,
        ],
    )

    return parser


def set_cfg(argv):
    parser = get_parser()
    args = parser.parse_known_args(argv)

    for n, v in vars(args[0]).items():
        setattr(cfg, n, v)

    # trick for showing ALL Stacks
    # if nor stack nor role nor type are specified.
    if cfg.all_stacks and not (cfg.stack or cfg.role or cfg.type):
        cfg.type = ["ALL"]

    if not cfg.no_stacks and not (cfg.stack or cfg.role or cfg.type):
        parser.print_help()
        exit(0)

    cfg.stack_args = args[1]



import boto3
from . import cfg


class myboto3(object):
    def __init__(self, istack=None, region=None):
        self.istack = istack

        try:
            cfg.parallel
            self.parallel = cfg.parallel
        except Exception:
            self.parallel = None

        kwarg_session = {}
        region_name = region if region else cfg.region
        if region_name:
            kwarg_session["region_name"] = region_name

        if cfg.profile:
            kwarg_session["profile_name"] = cfg.profile

        if not self.parallel and not region:
            try:
                self.boto3 = cfg.boto3
            except Exception:
                self.boto3 = boto3.session.Session(**kwarg_session)
                cfg.boto3 = self.boto3
        else:
            self.boto3 = boto3.session.Session(**kwarg_session)

        self.region_name = self.boto3.region_name

    def client(self, name):
        attr_name = f"cli_{self.region_name}_{name}"

        if self.parallel and self.istack:
            obj = self.istack
        else:
            obj = cfg

        try:
            client = getattr(obj, attr_name)
        except Exception:
            client = self.boto3.client(name)
            setattr(obj, attr_name, client)

        return client

    def resource(self, name):
        attr_name = f"res_{self.region_name}_{name}"

        if self.parallel:
            obj = self.istack
        else:
            obj = cfg

        try:
            resource = getattr(obj, attr_name)
        except Exception:
            resource = self.boto3.resource(name)
            setattr(obj, attr_name, resource)

        return resource



import logging
from . import cfg
from .common import *

try:
    import slack

    HAVE_SLACK = True
except ModuleNotFoundError:
    HAVE_SLACK = False

logging.basicConfig()
logging.getLogger("botocore").setLevel("CRITICAL")
logger = logging.getLogger("stacksops")
logger.setLevel(logging.INFO)


def get_msg_client():
    try:
        cfg.slack_channel
    except Exception:
        cfg.slack_channel = None

    if (
        cfg.slack_channel
        and HAVE_SLACK
        and "IBOX_SLACK_TOKEN" in os.environ
        and "IBOX_SLACK_USER" in os.environ
    ):
        slack_web = slack.WebClient(token=os.environ["IBOX_SLACK_TOKEN"])
        return slack_web

    return None



from prettytable import PrettyTable, ALL as ptALL
from .tools import show_confirm
from .common import *


# create changeset
def _do_changeset(istack, us_args):
    if not istack.cfg.showtags:
        # keep existing stack.tags so that they are excluded from changeset
        # (not the new prepared ones)
        us_args["Tags"] = istack.stack.tags

    us_args["StackName"] = istack.name
    us_args["ChangeSetName"] = "CS-%s" % time.strftime("%Y-%m-%dT%H-%M-%S")
    us_args.pop("StackPolicyDuringUpdateBody", None)

    response = istack.client.create_change_set(**us_args)

    return response["Id"]


# wait until changeset is created
def _changeset_waiter(istack, changeset_id):
    while True:
        time.sleep(3)
        changeset = istack.client.describe_change_set(
            ChangeSetName=changeset_id, StackName=istack.name
        )
        if changeset["Status"] in istack.cfg.CHANGESET_COMPLETE_STATUS:
            return changeset


# parse changeset changes
def _parse_changeset(changeset):
    changes = []
    for change in changeset["Changes"]:
        change_dict = {}
        ResChange = change["ResourceChange"]
        change_dict["LogicalResourceId"] = ResChange["LogicalResourceId"]
        change_dict["Action"] = ResChange["Action"]
        change_dict["ResourceType"] = ResChange["ResourceType"]
        if ResChange["Action"] == "Modify":
            change_dict["Replacement"] = ResChange["Replacement"]
            scope = ResChange["Scope"]
            change_dict["Scope"] = list_to_string_list(scope)
            target = []
            causingentity = []
            for i in ResChange["Details"]:
                if "Name" in i["Target"] and i["Target"]["Name"] not in target:
                    target.append(i["Target"]["Name"])
                if (
                    "CausingEntity" in i
                    and "Name" in i["Target"]
                    and i["CausingEntity"] not in causingentity
                ):
                    causingentity.append(i["CausingEntity"])
            change_dict["Target"] = list_to_string_list(target)
            change_dict["CausingEntity"] = list_to_string_list(causingentity)

        changes.append(change_dict)
    return changes


def _show_changeset_changes(istack, changes):
    fields = ["LogicalResourceId", "ResourceType", "Action"]
    fileds_ex = ["Replacement", "Scope", "Target", "CausingEntity"]
    fields.extend(fileds_ex)
    table = PrettyTable()
    if istack.cfg.nodetails:
        fields.remove("Target")
        fields.remove("CausingEntity")
    table.field_names = fields
    table.padding_width = 1
    table.align["LogicalResourceId"] = "l"
    table.align["ResourceType"] = "l"
    for row in changes:
        table.add_row(
            [
                "None" if i in fileds_ex and row["Action"] != "Modify" else row[i]
                for i in fields
            ]
        )

    istack.mylog("ChangeSet:")
    print(table.get_string(fields=fields))


def _delete_changeset(istack, changeset_id):
    response = istack.client.delete_change_set(
        ChangeSetName=changeset_id, StackName=istack.name
    )

    return response


def _execute_changeset(istack, changeset_id):
    response = istack.client.execute_change_set(
        ChangeSetName=changeset_id, StackName=istack.name
    )

    return response


# change list in a string new line separated element
def list_to_string_list(mylist):
    joined_string = "\n".join(mylist)
    mystring = joined_string if len(mylist) < 2 else f"({joined_string})"

    return mystring


def process(istack, us_args):
    # -create changeset
    changeset_id = _do_changeset(istack, us_args.copy())
    print("\n")
    istack.mylog("ChangeSetId: %s" % changeset_id)
    print("\n")
    time.sleep(1)
    istack.mylog("Waiting ChangeSet Creation..")

    # -wait changeset creation and return it
    changeset = _changeset_waiter(istack, changeset_id)
    # pprint(changeset)

    # -parse changeset changes
    changeset_changes = _parse_changeset(changeset)

    # -show changeset changes
    _show_changeset_changes(istack, changeset_changes)

    # -delete changeset
    _delete_changeset(istack, changeset_id)

    if not istack.cfg.dryrun and show_confirm():
        # _execute_changeset(istack, changeset_id)
        return True
    else:
        return None
        # _delete_changeset(istack, changeset_id)



import argparse
from .log import logger
from .common import *


# set final parameters values to use for exectuing commands -
# istack.action_parameters and istack.r_parameters
def _set_action_parameters(
    istack, params_default, params_changed, params_added, params_forced_default
):
    for key in sorted(istack.parameters):
        v = istack.parameters[key]

        try:
            default_value = v["Default"]
        except Exception:
            default_value = None
        use_previous_value = False

        # get list of AllowedValues
        allowed_values = v["AllowedValues"] if "AllowedValues" in v else []

        # check if key exist as istack.cfg param/attr too
        try:
            cfg_value = getattr(istack.cfg, key)
            in_cfg = True if cfg_value is not None else None
        except Exception:
            in_cfg = None

        # update param is not new ...
        if key in istack.c_parameters:
            current_value = istack.c_parameters[key]

            # current value its different from specified cmd arg
            if in_cfg and current_value != cfg_value:
                # get value from specified cmd arg
                value = cfg_value
                params_changed[key] = f"{current_value} => {value}"

            # current value is not allowed by new template
            elif len(allowed_values) > 0 and (current_value not in allowed_values):
                # get value from template default
                value = default_value
                params_forced_default[key] = f"{current_value} => {default_value}"

            # current value is unchanged and allowed
            else:
                value = ""
                use_previous_value = True
                params_default[key] = current_value

        # update param is new ...
        else:
            # template default value its different from specified cmd arg
            if in_cfg and default_value != cfg_value:
                value = cfg_value
                params_changed[key] = value

            # no cmd arg for new param
            # should never be here make a change to enforce param
            # in add_stack_params_as_args
            else:
                # get template default value
                value = default_value
                params_added[key] = default_value

        # append dictionary element to list
        istack.action_parameters.append(
            {
                "ParameterKey": key,
                "ParameterValue": str(value),
                "UsePreviousValue": use_previous_value,
            }
            if istack.stack
            else {
                "ParameterKey": key,
                "ParameterValue": str(value),
            }
        )

        # update resolved parameter final value istack.r_parameters
        istack.r_parameters[key] = current_value if use_previous_value else value


# force EnvShort param value based on Env one
def _force_envshort(istack):
    # use arg if exist or use current value
    env = istack.cfg.Env if istack.cfg.Env else istack.c_parameters["Env"]

    env_envshort_dict = {
        "dev": "dev",
        "stg": "stg",
        "prd": "prd",
        "prod": "prd",
    }

    istack.cfg.EnvShort = env_envshort_dict[env]


# for template in s3, try to get version from url/key otherway use fixed value "1"
# Ex for version=master-2ed25d5:
# https://eu-west-1-ibox-app-repository.s3.amazonaws.com/ibox/master-2ed25d5/templates/cache-portal.json
def _do_envstackversion_from_s3_template(istack):
    template = istack.cfg.template
    istack.cfg.EnvStackVersion = "1"

    if str(template).startswith("https"):
        try:
            istack.cfg.EnvStackVersion = template.split("/")[4]
        except Exception:
            pass


def get_stack_parameter_parser(istack):
    parser = argparse.ArgumentParser(
        description="",
        add_help=False,
        allow_abbrev=False,
        usage="Allowed Stack Params ... allowed values are in {}",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # for parameter in sorted(
    #    stack_parameters, key=lambda x: x['ParameterKey']):
    for p in sorted(istack.parameters):
        v = istack.parameters[p]
        allowed_values = v["AllowedValues"] if "AllowedValues" in v else []
        kwargs = {"type": str, "metavar": "\t%s" % v["Description"]}

        # If Parameter do not have Default value and is new or
        # current value is not allowed in new template,
        # enforce it as required
        if "Default" not in v and (
            p not in istack.c_parameters
            or (allowed_values and istack.c_parameters[p] not in allowed_values)
        ):
            kwargs["required"] = True

        if len(allowed_values) > 0:
            kwargs["choices"] = allowed_values
            kwargs["help"] = "{%s}\n\n" % ", ".join(allowed_values)
        else:
            kwargs["help"] = "\n"

        parser.add_argument("--%s" % p, **kwargs)

    return parser


def add_stack_params_as_args(istack, parser):
    args = parser.parse_args(istack.cfg.stack_args)
    istack.stack_parsed_args = args

    for n, v in vars(args).items():
        if not hasattr(istack.cfg, n):
            setattr(istack.cfg, n, v)


def show_override(istack):
    params = {}
    for name, value in istack.c_parameters.items():
        if (
            not name.startswith("Env")
            and any(name not in n for n in ["UpdateMode"])
            and any(
                name == t_name and (value != t_value.get("Default"))
                for t_name, t_value in istack.parameters.items()
            )
        ):
            params[name] = value

    out = pformat(params, width=istack.cfg.OUT_WIDTH)

    istack.mylog(f"CURRENT NOT DEFAULT - STACK PARAMETERS\n{out}\n")


def process(istack, show=True):
    logger.info("Processing Parameters")

    # get stack parameter parser
    parser = get_stack_parameter_parser(istack)

    # add stack parameters as argparse args and update istack.cfg
    add_stack_params_as_args(istack, parser)

    # if template include EnvShort params force its value based on the Env one
    if "EnvShort" in istack.parameters:
        _force_envshort(istack)

    # if using template option set/force EnvStackVersion
    if istack.cfg.template:
        _do_envstackversion_from_s3_template(istack)

    # unchanged stack params
    params_default = {}

    # changed stack params
    params_changed = {}

    # new stack params - default value
    params_added = {}

    # forced to default stack params - current value not in allowed ones
    params_forced_default = {}

    # list of final parameters args to use for executing action
    # as dict with ParameterKey ParameterValue keys
    # Ex for EnvRole=cache:
    # [{u'ParameterKey': 'EnvRole', u'ParameterValue': 'cache'}, {...}]
    istack.action_parameters = []

    # final resolved value stack parameters - {name: value} dictionary
    istack.r_parameters = {}

    # set final parameters values to use for exectuing action -
    # istack.action_parameters and istack.r_parameters
    _set_action_parameters(
        istack, params_default, params_changed, params_added, params_forced_default
    )

    if not show:
        return

    # show changes to output
    print("\n")
    # disabled for now - it make no sense to display them
    # if not istack.stack and params_default:
    #     print('DEFAULT - STACK PARAMETERS\n%s\n' % pformat(
    #         params_default, width=1000000))

    if params_changed:
        istack.mylog(
            "CHANGED - STACK PARAMETERS\n%s\n" % pformat(params_changed, width=1000000)
        )

    if istack.stack and params_added:
        istack.mylog(
            "ADDED - STACK PARAMETERS\n%s\n" % pformat(params_added, width=1000000)
        )

    if params_forced_default:
        istack.mylog(
            "FORCED TO DEFAULT - STACK PARAMETERS\n%s\n"
            % pformat(params_forced_default, width=1000000)
        )


def get(stack):
    try:
        s_parameters = stack["Parameters"]
    except Exception as e:
        pass
    else:
        parameters = {}
        for parameter in s_parameters:
            key = parameter["ParameterKey"]
            # Think is better to use always value not the resolved one,
            # or i will have problem later when i check if value is allowed
            # value = parameter.get(
            #     'ResolvedValue', parameter.get('ParameterValue'))
            value = parameter.get("ParameterValue")
            parameters[key] = value

        return parameters



from collections import OrderedDict
from .log import logger, get_msg_client
from .common import *


def _update_template_param(istack):
    # try to get role from istack.cfg or use current stack parameter value
    try:
        role = istack.cfg.EnvRole
    except Exception:
        role = istack.c_parameters["EnvRole"]

    app_repository = istack.exports["BucketAppRepository"]
    s3_prefix = "ibox/%s/templates/%s." % (istack.cfg.version, role)

    try:
        response = istack.s3.list_objects_v2(Bucket=app_repository, Prefix=s3_prefix)
        istack.cfg.template = "https://%s.s3.amazonaws.com/%s" % (
            app_repository,
            response["Contents"][0]["Key"],
        )
    except Exception:
        raise IboxError(f"Error retrieving stack template with prefix: {s3_prefix}")


def get_template(istack, stackset=None):
    logger.info("Getting Template Body")
    # update template param if using version one
    if istack.cfg.version:
        _update_template_param(istack)

    try:
        # New template
        if istack.cfg.template:
            template = str(istack.cfg.template)
            # From s3
            if template.startswith("https"):
                url = template[template.find("//") + 2 :]
                bucket = url[: url.find(".")]
                key = url[url.find("/") + 1 :]

                response = istack.s3.get_object(Bucket=bucket, Key=key)
                body = response["Body"].read()
                istack.template_from = "S3"
            # From file
            else:
                f = open(istack.cfg.template, "r")
                body = f.read()
                istack.template_from = "File"
        # Current template
        else:
            if stackset:
                tbody = istack.TemplateBody
            else:
                response = istack.client.get_template(StackName=istack.name)
                tbody = response["TemplateBody"]
            if isinstance(tbody, OrderedDict):
                body = json.dumps(tbody)
            else:
                body = tbody
            istack.template_from = "Current"

    except Exception as e:
        raise IboxError(f"Error retrieving template: {e}")
    else:
        try:
            template = json.loads(body)
        except Exception:
            try:
                template = yaml.load(body, Loader=yaml.FullLoader)
            except Exception:
                raise IboxError("Error parsing template body")

        for n in ["Parameters", "Conditions", "Mappings", "Resources"]:
            setattr(istack, n.lower(), template.get(n))

        return template



from collections import OrderedDict
from .log import logger
from .common import *


def _process_template(istack):
    def _recursive_resolve(name, value):
        if isinstance(value, (dict, OrderedDict)):
            r_root = {}
            for r, v in value.items():
                if r == "Fn::If":
                    return _resolve_if(name, v)
                elif r == "Ref":
                    return _resolve_ref(name, v)
                elif r == "Fn::GetAtt":
                    return "!GetAtt %s.%s" % (v[0], v[1])
                elif r == "Fn::GetAZs":
                    return "!GetAZs %s" % istack.boto3.region_name
                elif r == "Fn::ImportValue":
                    return _resolve_import(name, v)
                elif r == "Fn::Sub":
                    return _resolve_sub(name, v)
                elif r == "Fn::FindInMap":
                    return _resolve_findinmap(name, v)
                elif r == "Fn::Join":
                    return _resolve_join(name, v)
                elif r == "Fn::Select":
                    return _resolve_select(name, v)
                elif r == "Fn::Split":
                    return _resolve_split(name, v)
                elif r == "Fn::Or":
                    return _resolve_or(name, v)
                elif r == "Fn::And":
                    return _resolve_and(name, v)
                elif r == "Fn::Equals":
                    return _resolve_equals(name, v)
                elif r == "Fn::Not":
                    return _resolve_not(name, v)
                elif r == "Condition" and isinstance(v, str):
                    return _resolve_condition(name, v)
                elif r == "Fn::Base64":
                    return _recursive_resolve(name, v)
                else:
                    r_value = _recursive_resolve(name, v)

                    if not _awsnovalue(r_value):
                        r_root[r] = r_value

            return r_root

        elif isinstance(value, list):
            r_root = []
            for n, l in enumerate(value):
                r_value = _recursive_resolve(n, l)

                if not _awsnovalue(r_value):
                    r_root.append(r_value)

            return r_root

        elif isinstance(value, (int, str)):

            return value

    def _resolve_sub(name, value):
        if isinstance(value, list):
            sub_string = value[0]
            sub_data = value[1]
        else:
            sub_string = value
            sub_data = ""

        while True:
            found = sub_string.find("${")
            if found == -1:
                break
            find_start = found + 2
            find_end = sub_string[find_start:].find("}") + find_start
            key = sub_string[find_start:find_end]
            replace_from = "${" + key + "}"

            if key in sub_data:
                r_value = _recursive_resolve(key, sub_data[key])
                replace_to = r_value
            elif key in istack.r_parameters:
                replace_to = istack.r_parameters[key]
            elif key == "AWS::Region":
                replace_to = istack.boto3.region_name
            elif key == "AWS::AccountId":
                replace_to = istack.boto3.client("sts").get_caller_identity()["Account"]
            else:
                replace_to = key

            sub_string = sub_string.replace(replace_from, str(replace_to))

        if sub_string.startswith("https://"):
            _find_s3_files(name, sub_string)

        return sub_string

    def _resolve_if(name, v):
        value = v[1] if istack.r_conditions[v[0]] else v[2]

        return _recursive_resolve(name, value)

    def _resolve_ref(name, v):
        if v == "AWS::Region":
            value = istack.boto3.region_name
        elif v in istack.r_parameters:
            value = istack.r_parameters[v]
        else:
            # value = {'Ref': v}
            value = f"!Ref {v}"

        return value

    def _resolve_import(name, v):
        value = _recursive_resolve(name, v)

        return istack.exports[value]

    def _resolve_findinmap(name, v):
        mapname = _recursive_resolve(name, v[0])
        key = _recursive_resolve(name, v[1])
        key_value = _recursive_resolve(name, v[2])

        value = istack.mappings[mapname][key][key_value]

        if isinstance(value, list):
            return value
        else:
            return str(value)

        return str(istack.mappings[mapname][key][key_value])

    def _resolve_join(name, v):
        sep = v[0]
        data = v[1]

        return sep.join(map(str, _recursive_resolve(name, data)))

    def _resolve_select(name, v):
        index = v[0]
        s_list = _recursive_resolve(name, v[1])

        try:
            value = s_list[index]
        except Exception:
            value = s_list

        return value

    def _resolve_split(name, v):
        delimeter = v[0]
        s_string = _recursive_resolve(name, v[1])

        return s_string.split(delimeter)

    def _resolve_or(name, v):
        o_list = []
        for n in v:
            o_list.append(_recursive_resolve(name, n))

        return any(o_list)

    def _resolve_and(name, v):
        o_list = []
        for n in v:
            o_list.append(_recursive_resolve(name, n))

        return all(o_list)

    def _resolve_equals(name, v):
        first_value = str(_recursive_resolve(name, v[0]))
        second_value = str(_recursive_resolve(name, v[1]))

        return True if first_value == second_value else False

    def _resolve_not(name, v):
        value = True if not _recursive_resolve(name, v[0]) else False

        return value

    def _resolve_condition(name, v):
        istack.r_conditions[v] = _recursive_resolve(name, istack.conditions[v])

        return istack.r_conditions[v]

    def _awsnovalue(value):
        if value == "AWS::NoValue" or value == "!Ref AWS::NoValue":
            return True

        return False

    def _find_s3_files(name, sub_string):
        if (
            "AWS::AutoScaling::LaunchConfiguration" in istack.t_resources
            and istack.t_resources["AWS::AutoScaling::LaunchConfiguration"] == name
        ):
            data = sub_string[8:].partition("/")
            host = data[0]
            if not host.endswith("amazonaws.com"):
                return
            key = data[2]
            s_bucket = host.rsplit(".", 3)
            if s_bucket[1].startswith("s3-"):
                bucket = s_bucket[0]
            else:
                bucket = host[: host.rfind(".s3.")]

            if host == bucket:
                return

            istack.s3_files.add((bucket, key))

    def _check_lambda(r):
        try:
            bucket = istack.r_resources[r]["Properties"]["Code"]["S3Bucket"]
            key = istack.r_resources[r]["Properties"]["Code"]["S3Key"]
        except Exception:
            pass
        else:
            istack.s3_files.add((bucket, key))

    def _process_conditions():
        istack.r_conditions = {}
        for c in sorted(istack.conditions):
            v = istack.conditions[c]
            if c not in istack.r_conditions:
                istack.r_conditions[c] = _recursive_resolve(c, v)

    def _process_resources():
        istack.r_resources = {}
        istack.t_resources = {}
        istack.s3_files = set()

        for r in sorted(istack.resources):
            v = istack.resources[r]
            if not ("Condition" in v and not istack.r_conditions[v["Condition"]]):
                try:
                    del v["Condition"]
                except Exception:
                    pass
                istack.t_resources[v["Type"]] = r
                istack.r_resources[r] = _recursive_resolve(r, v)

                if v["Type"] == "AWS::Lambda::Function":
                    _check_lambda(r)

    _process_conditions()
    _process_resources()


def _check_s3_files(istack):
    for f in istack.s3_files:
        bucket = f[0]
        key = f[1]
        try:
            istack.s3.head_object(Bucket=bucket, Key=key)
        except botocore.exceptions.ClientError as e:
            logger.error(f"Missing: {bucket}/{key}")
            raise IboxError(e)


def _check_ecr_images(istack):
    name = istack.t_resources["AWS::ECS::TaskDefinition"]
    images = []
    ecr = istack.boto3.client("ecr")

    for c in istack.r_resources[name]["Properties"]["ContainerDefinitions"]:
        image = c["Image"]
        registry_id = image[0 : image.find(".")]
        repository_name = image[image.find("/") + 1 : image.find(":")]
        image_id = image[image.find(":") + 1 :]
        # Skip already processed images and images from public docker repo
        if image not in images and ".dkr.ecr." in image:
            try:
                ecr.describe_images(
                    registryId=registry_id,
                    repositoryName=repository_name,
                    imageIds=[
                        {
                            "imageTag": image_id,
                        }
                    ],
                )
                images.append(image)
            except botocore.exceptions.ClientError as e:
                logger.error(f"Missing: {image}")
                raise IboxError(e)


def _do_check(istack):
    if istack.s3_files:
        _check_s3_files(istack)

    if "AWS::ECS::TaskDefinition" in istack.t_resources:
        _check_ecr_images(istack)


def show(istack):
    _process_template(istack)
    print("")
    logger.info(f"Resolved: {istack.name}")
    print(yaml.dump(istack.r_resources))


def process(istack):
    logger.debug("Processing Template")

    try:
        _process_template(istack)
        _do_check(istack)
    except IboxError:
        raise
    except Exception as e:
        pprint(e)
        logger.warning(
            "Error resolving template. "
            "Will not be able to validate s3 files and ecr images."
        )



def set_changed(istack):
    istack.after["resources"] = get(istack)

    before = istack.before["resources"]
    after = istack.after["resources"]

    changed = {}
    for r, v in before.items():
        if r in after and v != after[r]:
            changed[r] = after[r]

    istack.changed["resources"] = changed


def get(istack, dash=None):
    resources = {}
    res_list = list(istack.cfg.RESOURCES_MAP.keys())

    paginator = istack.client.get_paginator("list_stack_resources")
    response_iterator = paginator.paginate(StackName=istack.name)

    for r in response_iterator:
        for res in r["StackResourceSummaries"]:
            res_lid = res["LogicalResourceId"]
            res_type = res["ResourceType"]
            res_pid = res.get("PhysicalResourceId", "null")

            if res_lid in res_list:
                if res_pid.startswith("arn"):
                    res_pid = res_pid.split(":", 5)[5]
                if res_lid in [
                    "ListenerHttpsExternalRules1",
                    "ListenerHttpsExternalRules2",
                    "ListenerHttpExternalRules1",
                    "ListenerHttpInternalRules1",
                ]:
                    res_pid = "/".join(res_pid.split("/")[1:4])
                if res_lid == "ScalableTarget":
                    res_pid = res_pid.split("/")[1]
                if res_lid in ["Service", "ServiceSpot"]:
                    res_pid_arr = res_pid.split("/")
                    if len(res_pid_arr) == 3:
                        res_pid = res_pid_arr[2]
                    else:
                        res_pid = res_pid_arr[1]
                if res_lid in [
                    "LoadBalancerApplicationExternal",
                    "LoadBalancerApplicationInternal",
                ]:
                    res_pid = "/".join(res_pid.split("/")[1:4])

                if dash and istack.cfg.RESOURCES_MAP[res_lid]:
                    res_lid = istack.cfg.RESOURCES_MAP[res_lid]

                resources[res_lid] = res_pid
            elif res_type in res_list:
                # If RESOURCES_MAP directly include AWS Resource Type,
                # used for AWS::ServiceDiscovery::Service
                if res_pid.startswith("arn"):
                    res_pid = res_pid.split(":", 5)[5]
                resources[res_lid] = res_pid

    return resources



from prettytable import PrettyTable, ALL as ptALL
from . import cfg, i_stack
from .aws import myboto3
from .tools import concurrent_exec
from .log import logger
from .common import *


def _get_ssm_parameter(ssm, param):
    resp = ssm.get_parameter(Name=param)

    return resp["Parameter"]["Value"]


def get_setupped_regions():
    boto3 = myboto3()
    ssm = boto3.client("ssm")

    try:
        rgs = _get_ssm_parameter(ssm, f"{cfg.SSM_BASE_PATH}/regions")
    except Exception:
        return []

    return rgs.split()


def get_by_path(iregion, path):
    stacks_list = tuple(iregion.bdata.keys())
    params = {}

    paginator = iregion.ssm.get_paginator("get_parameters_by_path")
    response_iterator = paginator.paginate(Path=path, Recursive=True)

    for page in response_iterator:
        for p in page["Parameters"]:
            name = p["Name"]
            name = "/".join(name.split("/")[-2:])
            value = p["Value"]

            if "ALL" in iregion.cfg.type or name.startswith(stacks_list):
                params[name] = value

    return params


def put_parameters(iobj, params):
    for p in params:
        resp = iobj.ssm.put_parameter(
            Name=p["name"],
            Description=p["desc"],
            Value=p["value"],
            Type="String",
            Overwrite=True,
            Tier="Standard",
        )


def setup(iregion):
    param = {
        "name": f"{cfg.SSM_BASE_PATH}/regions",
        "desc": "Regions where to replicate",
        "value": " ".join(cfg.regions),
    }

    result = {}
    if len(iregion.bdata) == 0:
        put_parameters(iregion, [param])
        result = cfg.regions
    else:
        stack_data = {}
        for n, _ in iregion.bdata.items():
            s_param = dict(param)
            s_param["name"] = f"{cfg.SSM_BASE_PATH}/{n}/regions"
            stack_data[n] = [s_param]

        result = concurrent_exec("ssm", stack_data, i_stack, region=iregion.name)

    return result


def put(iregion):
    stacks_data = {}
    for s, v in iregion.bdata.items():
        stack = s
        args = v[0]
        parameters = v[1]
        ssm_params = []
        for p, v in vars(args).items():
            if not v:
                continue
            s_param = {
                "name": f"{cfg.SSM_BASE_PATH}/{s}/{p}",
                "desc": parameters[p]["Description"],
                "value": v,
            }
            ssm_params.append(s_param)
        stacks_data[s] = ssm_params

    result = concurrent_exec("ssm", stacks_data, i_stack, region=iregion.name)

    return result


def show(data):
    params_map = {}
    params_keys = []
    table = PrettyTable()
    table.padding_width = 1

    for r, v in data.items():
        params_map[r] = v
        params_keys.extend(list(v.keys()))

    params_keys = sorted(list(set(params_keys)))
    table.add_column("Parameter", params_keys)

    for r, v in params_map.items():
        params_values = []
        for n in params_keys:
            if n in v:
                params_values.append(v[n])
            else:
                params_values.append("")
        table.add_column(r, params_values)

    table.align["Parameter"] = "l"

    return table



from . import resources
from .common import *

from copy import deepcopy


def add_stack(istack):
    widget_width = {
        "stack": 12,
        "global": 24,
    }

    widget_height = {
        "stack": 6,
        "global": 6,
    }

    widget_title = {
        "5xx": "5xx - 4xx",
        "req": "Requests - Healthy",
        "net": "NetworkIN - NetworkOUT",
        "5xx_elb": "5xx - 4xx [ELB]",
        "50x_elb": "50x External - 50x Internal [ELB]",
    }

    widget_map = {
        "role": ["cpu", "response"],
        "req": ["requests", "healthy"],
        "5xx": ["5xx", "4xx"],
        "5xx_elb": ["5xx_elb", "4xx_elb"],
        "50x_elb": ["500_elb", "502_elb", "503_elb", "504_elb"],
        "net": ["netin", "netout"],
    }

    ScalingPolicyTrackingsCpuBaseLabel = "ScalingPolicyTrackingsCpu"
    widget_label = {}

    def get_alarm(res):
        alarms = {}
        cloudwatch = istack.boto3.resource("cloudwatch")
        for a in ["AlarmCPUHigh", "AlarmCPULow"]:
            alarm = cloudwatch.Alarm(res[a])
            alarms[a] = alarm.threshold

        return alarms["AlarmCPUHigh"], alarms["AlarmCPULow"]

    def get_policy_ec2(res):
        polname = res["ScalingPolicyTrackings1"].split("/")[2]
        AutoScalingGroupName = (
            "AutoScalingGroupSpotName"
            if "AutoScalingGroupSpotName" in res
            else "AutoScalingGroupName"
        )
        client = istack.boto3.client("autoscaling")
        response = client.describe_policies(
            AutoScalingGroupName=res[AutoScalingGroupName],
            PolicyNames=[polname],
        )

        return response["ScalingPolicies"][0]["TargetTrackingConfiguration"]

    def get_policy_ecs(res):
        resname = "/".join(res["ScalingPolicyTrackingsApp"].split("/")[2:5]).split(":")[
            0
        ]
        client = istack.boto3.client("application-autoscaling")
        response = client.describe_scaling_policies(
            PolicyNames=list(istack.cfg.SCALING_POLICY_TRACKINGS_NAMES.keys()),
            ResourceId=resname,
            ServiceNamespace="ecs",
        )

        return response["ScalingPolicies"][0][
            "TargetTrackingScalingPolicyConfiguration"
        ]

    def get_policy(res):
        ret = []
        for n in ["AutoScalingGroupName", "ServiceName"]:
            if n not in res:
                continue
            if n == "AutoScalingGroupName":
                conf = get_policy_ec2(res)
                l_type = "EC2"
                color = "#1f77b4"
            if n == "ServiceName":
                conf = get_policy_ecs(res)
                l_type = "ECS"
                color = "#36dc52"

            stat = "Average"
            value = conf["TargetValue"]
            if (
                "CustomizedMetricSpecification" in conf
                and "Statistic" in conf["CustomizedMetricSpecification"]
            ):
                stat = conf["CustomizedMetricSpecification"]["Statistic"]

            ret.append(
                (
                    value,
                    f"{ScalingPolicyTrackingsCpuBaseLabel}{l_type}{stat}",
                    color,
                )
            )

        return ret

    def resolve_widget_map(name):
        count = 0
        for n in widget_map[name]:
            count += len(metrics[n])

        if count > 0:
            return True

    def get_widget_annotations(res):
        AlarmCPUHighThreshold = 60
        AlarmCPULowThreshold = 30
        ScalingPolicyTrackingsCpuValue = 80
        ScalingPolicyTrackingsCpuLabel = ScalingPolicyTrackingsCpuBaseLabel
        widget_annotations_type = None
        tracking_list = []

        if "AlarmCPUHigh" and "AlarmCPULow" in res:
            AlarmCPUHighThreshold, AlarmCPULowThreshold = get_alarm(res)
            widget_annotations_type = "step"

        if any(k in res for k in istack.cfg.SCALING_POLICY_TRACKINGS_NAMES):
            widget_annotations_type = "tracking"
            tracking_list.append({"value": 100})
            for n in get_policy(res):
                ScalingPolicyTrackingsCpuValue = n[0]
                ScalingPolicyTrackingsCpuLabel = n[1]
                ScalingPolicyTrackingsCpuColor = n[2]
                tracking_list.append(
                    {
                        "label": ScalingPolicyTrackingsCpuLabel,
                        "value": ScalingPolicyTrackingsCpuValue,
                        "color": ScalingPolicyTrackingsCpuColor,
                    }
                )

        widget_annotations = {
            "tracking": tracking_list,
            "step": [
                {
                    "label": "AlarmCPUHighThreshold",
                    "value": AlarmCPUHighThreshold,
                },
                {
                    "label": "AlarmCPULowThreshold",
                    "value": AlarmCPULowThreshold,
                    "color": "#2ca02c",
                },
            ],
        }

        return widget_annotations, widget_annotations_type

    def do_label_exist(w_label, w_metrics):
        for index, metric in enumerate(w_metrics):
            if isinstance(metric, dict) and w_label in list(metric.values()):
                return True
            for m in metric:
                if isinstance(m, dict) and w_label in list(m.values()):
                    return index

        return None

    def do_insert_metrics(m, widget):
        label = m["label"]
        metric = m["metric"]
        msg = m["name"]
        widget_metrics = widget["properties"]["metrics"]
        if do_label_exist(label, metric):
            label_index = do_label_exist(label, widget_metrics)
            if label_index is None:
                widget_metrics.append(metric)
                out_msg = "Added"
            else:
                widget_metrics[label_index] = metric
                out_msg = "-- Updated"
            if not istack.cfg.silent:
                print(f"\tMetrics: {msg} {out_msg}")

    def get_widget_map_position(wtype, index):
        if wtype == "stack":
            # even index (left side)
            if (index & 1) == 0:
                return (0, widget_height["stack"] * index)
            # odd index (right side)
            else:
                return (widget_width["stack"], widget_height["stack"] * index)
        # global type
        else:
            return (0, widget_height["global"] * index)

    def add_annotations(w, atype):
        if "annotations" not in w["properties"]:
            w["properties"]["annotations"] = {}

        annotations = w["properties"]["annotations"]
        # Vertical
        if atype == "vertical":
            value_now = datetime.utcnow().strftime("%Y-%m-%dT%X.000Z")
            w_ann_vertical = {
                "label": "",
                "value": value_now,
            }
            if istack.cfg.vertical in ["after", "before"]:
                w_ann_vertical["fill"] = istack.cfg.vertical

            # vertical annotation not present
            if "vertical" not in annotations:
                annotations["vertical"] = []
            # append only if empty or there is not annotations
            # with value in the same minute
            if not annotations["vertical"] or all(
                value_now[0:16] not in a["value"] for a in annotations["vertical"]
            ):
                annotations["vertical"].append(w_ann_vertical)

        # Horizontal annotations
        annotations["horizontal"] = widget_annotations.get(widget_annotations_type, [])

        # Horizontal (ec2-ecs)
        # if atype == 'ecs':
        #    annotations['horizontal'] = widget_annotations['ecs']
        # if atype == 'ec2':
        #    annotations['horizontal'] = widget_annotations['ec2']

    def get_widget_base(wtype, wlist, windex, title, w):
        widget = {}
        widget.update(
            {
                "type": "metric",
                "x": get_widget_map_position(wtype, windex)[0],
                "y": get_widget_map_position(wtype, windex)[1],
                "width": widget_width[wtype],
                "height": widget_height[wtype],
                "properties": {
                    "view": "timeSeries",
                    "stacked": False,
                    "metrics": [],
                    "region": istack.boto3.region_name,
                    "title": title,
                    "period": 300,
                },
            }
        )

        # If widget already exist get metrics from current one
        if len(wlist) > 0:
            widget["properties"]["metrics"] = w[windex]["properties"]["metrics"]
            # and if exists, annotations too..
            if "annotations" in w[windex]["properties"]:
                widget["properties"]["annotations"] = w[windex]["properties"][
                    "annotations"
                ]
            del w[windex]
            out_msg = "Updated"
        else:
            out_msg = "Added"

        if istack.cfg.vertical:
            add_annotations(widget, "vertical")

        w.insert(windex, widget)
        if not istack.cfg.silent:
            print(f"Widget:{title} {out_msg}")

        return widget

    def update_dashboard(res, dashboard_name):
        cw = istack.boto3.client("cloudwatch")

        if not istack.cfg.silent:
            print(dashboard_name)

        try:
            dashboard_body = cw.get_dashboard(DashboardName=dashboard_name)[
                "DashboardBody"
            ]
            dash = json.loads(dashboard_body)
        except Exception as e:
            print("DashBoard do not exist, creating one..\n")
            dash = {"widgets": []}

        w = dash["widgets"]

        # Find the current number of widget stacks,
        # so that the next one is added at the end
        len_stacks = len([n for n in w if n["width"] == widget_width["stack"]])

        # iterate over widget_map keys and populate relative widget
        for wd in widget_map:
            w_list = [
                n
                for n, v in enumerate(w)
                if v["properties"]["title"] == widget_title[wd]
            ]

            if wd == "role":
                w_index = len_stacks
                w_type = "stack"
            else:
                w_index = len(w)
                w_type = "global"

            if len(w_list) > 0:
                w_index = w_list[0]

            if resolve_widget_map(wd):
                # if relative metrics exists
                widget = get_widget_base(w_type, w_list, w_index, widget_title[wd], w)
                if wd == "role":
                    # add horizonatal annotations for role widget
                    add_annotations(widget, None)
                for n in widget_map[wd]:
                    for m in metrics[n]:
                        do_insert_metrics(m, widget)
        # END Widgets

        if istack.cfg.debug:
            print(json.dumps(dash, indent=4))
            return

        # Put DashBoard
        out = cw.put_dashboard(
            DashboardName=dashboard_name,
            DashboardBody=json.dumps(dash, separators=(",", ":")),
        )

        if len(out["DashboardValidationMessages"]) > 0:
            pprint(out)
        else:
            print("")
            mylog("CloudWatch-DashBoard[" + dashboard_name + "] Updated:")

        return True

    def mylog(string):
        print(istack.name + " # " + string)

    def add_spot_metric(metric, index):
        label = metric[index]["label"]
        stat = metric[index]["metric"][6]["stat"]
        metric.insert(
            index,
            {
                "name": metric[index]["name"],
                "label": label,
                "metric": [
                    {
                        "visible": True,
                        "expression": f"cpu{stat}Base + cpu{stat}Spot",
                        "label": label,
                        "id": f"cpu{stat}sum",
                    }
                ],
            },
        )
        i = index + 1
        metric.insert(i + 1, deepcopy(metric[i]))

        metric[i]["label"] += " - Base"
        metric[i]["name"] += " - Base"
        metric[i]["metric"][6].update(
            {"visible": False, "id": f"cpu{stat}Base", "label": metric[i]["label"]}
        )

        metric[i + 1]["label"] += " - Spot"
        metric[i + 1]["name"] += " - Spot"
        metric[i + 1]["metric"][6].update(
            {"visible": False, "id": f"cpu{stat}Spot", "label": metric[i + 1]["label"]}
        )
        metric[i + 1]["metric"][3] = res["ServiceSpotName"]

    def get_metrics(res):
        # update widget_title and widget_label
        widget_title["role"] = f"{istack.EnvRole}.{istack.name}"
        title_role = widget_title["role"]

        # Set common variable for ELB Classic and Application used by EC2 stack
        if any(
            n in res for n in ["LoadBalancerNameExternal", "LoadBalancerNameInternal"]
        ):
            LoadBalancerName = "LoadBalancerName"
            LoadBalancerNameExternal = "LoadBalancerNameExternal"
            LoadBalancerNameInternal = "LoadBalancerNameInternal"
            AWS_ELB = "AWS/ELB"
            Latency = "Latency"
            HTTPCode_Backend_5XX = "HTTPCode_Backend_5XX"
            HTTPCode_Backend_4XX = "HTTPCode_Backend_4XX"
            HTTPCode_ELB_5XX = "HTTPCode_ELB_5XX"
            HTTPCode_ELB_4XX = "HTTPCode_ELB_4XX"
        elif any(n in res for n in ["LoadBalancerExternal", "LoadBalancerInternal"]):
            LoadBalancerName = "LoadBalancer"
            LoadBalancerNameExternal = "LoadBalancerExternal"
            LoadBalancerNameInternal = "LoadBalancerInternal"
            AWS_ELB = "AWS/ApplicationELB"
            Latency = "TargetResponseTime"
            HTTPCode_Backend_5XX = "HTTPCode_Target_5XX_Count"
            HTTPCode_Backend_4XX = "HTTPCode_Target_4XX_Count"
            HTTPCode_ELB_5XX = "HTTPCode_ELB_5XX_Count"
            HTTPCode_ELB_4XX = "HTTPCode_ELB_4XX_Count"
        else:
            LoadBalancerName = None
            LoadBalancerNameExternal = None
            LoadBalancerNameInternal = None
            AWS_ELB = None
            Latency = None
            HTTPCode_Backend_5XX = None
            HTTPCode_Backend_4XX = None

        # build empty metrics dict from widget_map
        metrics = {n: [] for m in widget_map.values() for n in m}

        # Is a MIXED stack with both ECS Service and EC2 ASG?
        if all(n in res for n in ["ServiceName", "AutoScalingGroupName"]):
            IS_MIXED = True
        else:
            IS_MIXED = False

        # ECS
        if all(n in res for n in ["ServiceName", "ClusterName"]):
            L_TYPE = " ECS" if IS_MIXED else ""
            # CPU
            label = f"Cpu{L_TYPE} - {istack.cfg.statistic}"
            metrics["cpu"].append(
                {
                    "name": f"Cpu{L_TYPE}",
                    "label": label,
                    "metric": [
                        "AWS/ECS",
                        "CPUUtilization",
                        "ServiceName",
                        res["ServiceName"],
                        "ClusterName",
                        res["ClusterName"],
                        {
                            "period": 300,
                            "stat": istack.cfg.statistic,
                            "label": label,
                        },
                    ],
                }
            )
            # Always add cpu maximum
            label = f"Cpu{L_TYPE} - Maximum"
            metrics["cpu"].append(
                {
                    "name": f"Cpu{L_TYPE} - Maximum",
                    "label": label,
                    "metric": [
                        "AWS/ECS",
                        "CPUUtilization",
                        "ServiceName",
                        res["ServiceName"],
                        "ClusterName",
                        res["ClusterName"],
                        {
                            "period": 300,
                            "stat": "Maximum",
                            "label": label,
                        },
                    ],
                }
            )

            if "ServiceSpotName" in res:
                for n, v in enumerate(list(metrics["cpu"])):
                    add_spot_metric(metrics["cpu"], n * 3)

            # TargetGroups
            for n, v in res.items():
                if n.startswith("TargetGroup") and n.endswith(("External", "Internal")):
                    if n.endswith("External"):
                        suffix = "External"
                    else:
                        suffix = "Internal"

                    tg_name = n.replace("TargetGroup", "")

                    if f"LoadBalancer{suffix}" not in res:
                        continue
                    else:
                        lb_name = f"LoadBalancer{suffix}"

                    # Response Time
                    label = f"Response {tg_name}" f" - {istack.cfg.statisticresponse}"
                    metrics["response"].append(
                        {
                            "name": f"Response {tg_name}",
                            "label": label,
                            "metric": [
                                "AWS/ApplicationELB",
                                "TargetResponseTime",
                                "TargetGroup",
                                res[n],
                                "LoadBalancer",
                                res[lb_name],
                                {
                                    "period": 300,
                                    "stat": istack.cfg.statisticresponse,
                                    "yAxis": "right",
                                    "label": label,
                                },
                            ],
                        }
                    )
                    # Healthy
                    label = f"{title_role}{L_TYPE} - Healthy"
                    metrics["healthy"].append(
                        {
                            "name": f"Healthy{L_TYPE} {tg_name}",
                            "label": label,
                            "metric": [
                                "AWS/ApplicationELB",
                                "HealthyHostCount",
                                "TargetGroup",
                                res[n],
                                "LoadBalancer",
                                res[lb_name],
                                {
                                    "label": label,
                                    "stat": istack.cfg.statistic,
                                    "yAxis": "right",
                                },
                            ],
                        }
                    )
                    # Requests
                    label = f"{title_role} {tg_name} - Requests"
                    metrics["requests"].append(
                        {
                            "name": f"Requests {tg_name}",
                            "label": label,
                            "metric": [
                                "AWS/ApplicationELB",
                                "RequestCount",
                                "TargetGroup",
                                res[n],
                                "LoadBalancer",
                                res[lb_name],
                                {"label": label, "stat": "Sum"},
                            ],
                        }
                    )
                    # 5xx
                    label = f"{title_role} {tg_name} - 5xx"
                    metrics["5xx"].append(
                        {
                            "name": f"5xx {tg_name}",
                            "label": label,
                            "metric": [
                                "AWS/ApplicationELB",
                                "HTTPCode_Target_5XX_Count",
                                "TargetGroup",
                                res[n],
                                "LoadBalancer",
                                res[lb_name],
                                {"label": label, "stat": "Sum"},
                            ],
                        }
                    )
                    # 4xx
                    label = f"{title_role} {tg_name} - 4xx"
                    metrics["4xx"].append(
                        {
                            "name": f"4xx {tg_name}",
                            "label": label,
                            "metric": [
                                "AWS/ApplicationELB",
                                "HTTPCode_Target_4XX_Count",
                                "TargetGroup",
                                res[n],
                                "LoadBalancer",
                                res[lb_name],
                                {"label": label, "stat": "Sum", "yAxis": "right"},
                            ],
                        }
                    )
        if True:
            # EC2
            if all(n in res for n in ["AutoScalingGroupName"]):
                L_TYPE = " EC2" if IS_MIXED else ""
                # CPU
                label = f"Cpu{L_TYPE} - {istack.cfg.statistic}"
                metrics["cpu"].append(
                    {
                        "name": f"Cpu{L_TYPE}",
                        "label": label,
                        "metric": [
                            "AWS/EC2",
                            "CPUUtilization",
                            "AutoScalingGroupName",
                            res["AutoScalingGroupName"],
                            {
                                "period": 300,
                                "stat": istack.cfg.statistic,
                                "label": label,
                            },
                        ],
                    }
                )
                # Always add cpu maximum
                label = f"Cpu{L_TYPE} - Maximum"
                metrics["cpu"].append(
                    {
                        "name": f"Cpu{L_TYPE} - Maximum",
                        "label": label,
                        "metric": [
                            "AWS/EC2",
                            "CPUUtilization",
                            "AutoScalingGroupName",
                            res["AutoScalingGroupName"],
                            {
                                "period": 300,
                                "stat": "Maximum",
                                "label": label,
                            },
                        ],
                    }
                )
                # CPU Spot
                if all(n in res for n in ["AutoScalingGroupSpotName"]):
                    label = f"Cpu{L_TYPE} Spot - {istack.cfg.statistic}"
                    metrics["cpu"].append(
                        {
                            "name": f"Cpu{L_TYPE} Spot",
                            "label": label,
                            "metric": [
                                "AWS/EC2",
                                "CPUUtilization",
                                "AutoScalingGroupName",
                                res["AutoScalingGroupSpotName"],
                                {
                                    "period": 300,
                                    "stat": istack.cfg.statistic,
                                    "label": label,
                                },
                            ],
                        }
                    )
                # Healthy
                label = f"{title_role}{L_TYPE} - Healthy"
                metrics["healthy"].append(
                    {
                        "name": f"Healthy{L_TYPE}",
                        "label": label,
                        "metric": [
                            "AWS/AutoScaling",
                            "GroupInServiceInstances",
                            "AutoScalingGroupName",
                            res["AutoScalingGroupName"],
                            {
                                "label": label,
                                "stat": istack.cfg.statistic,
                                "yAxis": "right",
                            },
                        ],
                    }
                )
                # Network
                label = f"{title_role} - NetworkIN"
                metrics["netin"].append(
                    {
                        "name": "NetworkIN",
                        "label": label,
                        "metric": [
                            "AWS/EC2",
                            "NetworkIn",
                            "AutoScalingGroupName",
                            res["AutoScalingGroupName"],
                            {"label": label, "period": 300, "stat": "Sum"},
                        ],
                    }
                )
                label = f"{title_role} - NetworkOUT"
                metrics["netout"].append(
                    {
                        "name": "NetworkOUT",
                        "label": label,
                        "metric": [
                            "AWS/EC2",
                            "NetworkOut",
                            "AutoScalingGroupName",
                            res["AutoScalingGroupName"],
                            {
                                "label": label,
                                "period": 300,
                                "stat": "Sum",
                                "yAxis": "right",
                            },
                        ],
                    }
                )

            # ELB
            for n in ["External", "Internal"]:
                res_name = locals()[f"LoadBalancerName{n}"]

                if res_name in res and "ServiceName" not in res:
                    # Response
                    label = f"Response {n} - {istack.cfg.statisticresponse}"
                    metrics["response"].append(
                        {
                            "name": f"Response {n}",
                            "label": label,
                            "metric": [
                                AWS_ELB,
                                Latency,
                                LoadBalancerName,
                                res[res_name],
                                {
                                    "period": 300,
                                    "stat": istack.cfg.statisticresponse,
                                    "yAxis": "right",
                                    "label": label,
                                },
                            ],
                        }
                    )
                    # Requests
                    label = f"{title_role} {n} - Requests"
                    metrics["requests"].append(
                        {
                            "name": f"Requests {n}",
                            "label": label,
                            "metric": [
                                AWS_ELB,
                                "RequestCount",
                                LoadBalancerName,
                                res[res_name],
                                {"label": label, "stat": "Sum"},
                            ],
                        }
                    )
                    # 5xx
                    label = f"{title_role} {n} - 5xx"
                    metrics["5xx"].append(
                        {
                            "name": f"5xx {n}",
                            "label": label,
                            "metric": [
                                AWS_ELB,
                                HTTPCode_Backend_5XX,
                                LoadBalancerName,
                                res[res_name],
                                {"label": label, "stat": "Sum"},
                            ],
                        }
                    )
                    # 4xx
                    label = f"{title_role} {n} - 4xx"
                    metrics["4xx"].append(
                        {
                            "name": f"4xx {n}",
                            "label": label,
                            "metric": [
                                AWS_ELB,
                                HTTPCode_Backend_4XX,
                                LoadBalancerName,
                                res[res_name],
                                {"label": label, "stat": "Sum", "yAxis": "right"},
                            ],
                        }
                    )
                    # 5xx ELB
                    label = f"{title_role} {n} - 5xx"
                    metrics["5xx_elb"].append(
                        {
                            "name": f"5xx {n} ELB",
                            "label": label,
                            "metric": [
                                AWS_ELB,
                                HTTPCode_ELB_5XX,
                                LoadBalancerName,
                                res[res_name],
                                {"label": label, "stat": "Sum"},
                            ],
                        }
                    )
                    # 4xx ELB
                    label = f"{title_role} {n} - 4xx"
                    metrics["4xx_elb"].append(
                        {
                            "name": f"4xx {n} ELB",
                            "label": label,
                            "metric": [
                                AWS_ELB,
                                HTTPCode_ELB_4XX,
                                LoadBalancerName,
                                res[res_name],
                                {"label": label, "stat": "Sum", "yAxis": "right"},
                            ],
                        }
                    )

                    # 50x ELB
                    if f"LoadBalancer{n}" in res:
                        # 500
                        label = f"{title_role} {n} - 500"
                        metrics["500_elb"].append(
                            {
                                "name": f"500 {n} ELB",
                                "label": label,
                                "metric": [
                                    "AWS/ApplicationELB",
                                    "HTTPCode_ELB_500_Count",
                                    "LoadBalancer",
                                    res[res_name],
                                    {
                                        "label": label,
                                        "stat": "Sum",
                                        "yAxis": (
                                            "right" if n == "Internal" else "left"
                                        ),
                                    },
                                ],
                            }
                        )
                        # 502
                        label = f"{title_role} {n} - 502"
                        metrics["502_elb"].append(
                            {
                                "name": f"502 {n} ELB",
                                "label": label,
                                "metric": [
                                    "AWS/ApplicationELB",
                                    "HTTPCode_ELB_502_Count",
                                    "LoadBalancer",
                                    res[res_name],
                                    {
                                        "label": label,
                                        "stat": "Sum",
                                        "yAxis": (
                                            "right" if n == "Internal" else "left"
                                        ),
                                    },
                                ],
                            }
                        )
                        # 503
                        label = f"{title_role} {n} - 503"
                        metrics["503_elb"].append(
                            {
                                "name": f"503 {n} ELB",
                                "label": label,
                                "metric": [
                                    "AWS/ApplicationELB",
                                    "HTTPCode_ELB_503_Count",
                                    "LoadBalancer",
                                    res[res_name],
                                    {
                                        "label": label,
                                        "stat": "Sum",
                                        "yAxis": (
                                            "right" if n == "Internal" else "left"
                                        ),
                                    },
                                ],
                            }
                        )
                        # 504
                        label = f"{title_role} {n} - 504"
                        metrics["504_elb"].append(
                            {
                                "name": f"504 {n} ELB",
                                "label": label,
                                "metric": [
                                    "AWS/ApplicationELB",
                                    "HTTPCode_ELB_504_Count",
                                    "LoadBalancer",
                                    res[res_name],
                                    {
                                        "label": label,
                                        "stat": "Sum",
                                        "yAxis": (
                                            "right" if n == "Internal" else "left"
                                        ),
                                    },
                                ],
                            }
                        )

        return metrics

    # get stack resources
    res = resources.get(istack, dash=True)

    # get widget annotations and widget_annotations_type for alarms threshold
    # or policy tracking target value
    widget_annotations, widget_annotations_type = get_widget_annotations(res)

    if not istack.cfg.silent:
        pprint(res)

    # get metrics
    metrics = get_metrics(res)

    print("")
    # update dashboards
    update_dashboard(res, istack.cfg.dash_name)


def update(istack):
    cw_client = istack.boto3.client("cloudwatch")
    response_dash = cw_client.list_dashboards(DashboardNamePrefix="_")

    resources.set_changed(istack)
    if istack.cfg.dashboard == "OnChange":
        res_changed = istack.changed["resources"]
        mode = ""
    elif istack.cfg.dashboard in ["Always", "Generic"]:
        res_changed = istack.after["resources"]
        mode = istack.cfg.dashboard
    else:
        return

    if res_changed or any(
        n in istack.changed["outputs"]
        for n in [
            "ScalingPolicyTrackings",
            "AutoScalingScalingPolicy",
            "ApplicationAutoScalingScalingPolicy",
        ]
    ):
        for dash in response_dash["DashboardEntries"]:
            if istack.name in dash["DashboardName"]:
                istack.cfg.dash_name = dash["DashboardName"]
                add_stack(istack)



from . import cfg, parameters, outputs
from .aws import myboto3
from .log import logger
from .common import *


def get_base_data(stack, stackset=None):
    data = {
        "before": {},
        "after": {},
        "changed": {},
        "parameter_not_empty": [],
    }

    if stackset:
        data["TemplateBody"] = stack["TemplateBody"]
        data["Tags"] = stack["Tags"]
        data["StackSetId"] = stack["StackSetId"]

    stack_outputs = outputs.get(stack)

    data.update(stack_outputs)
    data["before"]["outputs"] = stack_outputs

    stack_parameters = parameters.get(stack)
    if stack_parameters:
        data["c_parameters"] = stack_parameters

        # add parameters too, this way i can show them using show command
        for n, v in data["c_parameters"].items():
            if (
                not n.startswith("Env")
                and n not in ["UpdateMode"]
                and v != ""
                and ",,," not in v
            ):
                data["parameter_not_empty"].append(n)
            if n not in data:
                data[n] = v

    return data


def _get_stack(r, data):
    for s in r["Stacks"]:
        stack_name = s["StackName"]
        stack_data = get_base_data(s)
        stack_role = stack_data.get("EnvRole", None)
        stack_type = stack_data.get("StackType", None)
        if (
            stack_name in cfg.stack
            or stack_role in cfg.role
            or stack_type in cfg.type
            or any(t in cfg.type for t in (stack_type.split() if stack_type else []))
            or "ALL" in cfg.type
        ):
            data[stack_name] = stack_data


def _get_stackset(r, data):
    s = r["StackSet"]
    stack_name = s["StackSetName"]
    stack_data = get_base_data(s, stackset=True)
    stack_role = stack_data.get("EnvRole", None)
    stack_type = stack_data.get("StackType", None)
    if (
        stack_name in cfg.stack
        or stack_role in cfg.role
        or stack_type in cfg.type
        or any(t in cfg.type for t in (stack_type.split() if stack_type else []))
        or "ALL" in cfg.type
    ):
        data[stack_name] = stack_data


def get(names=[], exit_if_empty=True, obj=None, stackset=None):
    if not obj:
        boto3 = myboto3()
        client = boto3.client("cloudformation")
    else:
        boto3 = getattr(obj, "boto3")
        client = boto3.client("cloudformation")

    logger.info("Getting Stacks Description")
    data = {}

    if not names:
        names = cfg.stack

    if stackset:
        response = client.describe_stack_set(StackSetName=cfg.stack[0])
        _get_stackset(response, data)
    elif not cfg.role and not cfg.type and len(names) < cfg.MAX_SINGLE_STACKS:
        for s in names:
            response = client.describe_stacks(StackName=s)
            _get_stack(response, data)
    else:
        paginator = client.get_paginator("describe_stacks")
        response_iterator = paginator.paginate()
        for r in response_iterator:
            _get_stack(r, data)

    if not data and exit_if_empty:
        logger.warning("No Stacks found!\n")
        exit(0)

    return data



from calendar import timegm
from .common import *


# show old and new service tasks during an update
def _show_service_update(istack, event, timedelta):
    # avoid showing current service log if is requested a stack past event
    if timedelta != "0":
        return

    service_logical_resource_id = event.logical_resource_id
    service = task = cluster = deps_before = None
    deployment_task = ""
    deployments_len = pendingCount = 0
    client = istack.boto3.client("ecs")

    try:
        cluster = istack.stack.Resource("ScalableTarget").physical_resource_id.split(
            "/"
        )[1]
        service = istack.stack.Resource(
            service_logical_resource_id
        ).physical_resource_id
    except Exception:
        return

    deps = {
        "PRIMARY": {},
        "ACTIVE": {},
        "INACTIVE": {},
        "DRAINING": {},
    }
    while task != deployment_task or deployments_len > 1 or pendingCount != 0:
        istack.stack.reload()
        task = istack.stack.Resource("TaskDefinition").physical_resource_id
        response = client.describe_services(
            cluster=cluster,
            services=[service],
        )
        deployments = response["services"][0]["deployments"]
        deployments_len = len(deployments)
        for dep in deployments:
            status = dep["status"]
            for p in [
                "taskDefinition",
                "desiredCount",
                "runningCount",
                "pendingCount",
                "failedTasks",
            ]:
                deps[status][p] = dep[p]

        deployment_task = deps["PRIMARY"]["taskDefinition"]
        desiredCount = deps["PRIMARY"]["desiredCount"]
        runningCount = deps["PRIMARY"]["runningCount"]
        pendingCount = deps["PRIMARY"]["pendingCount"]
        failedTasks = deps["PRIMARY"]["failedTasks"]

        if str(deps) != deps_before:
            deps_before = str(deps)
            for d in ["PRIMARY", "ACTIVE"]:
                if "taskDefinition" in deps[d]:
                    deps[d]["taskDefinition"] = deps[d]["taskDefinition"].split("/")[-1]
            istack.mylog("PRIMARY: %s" % pformat(deps["PRIMARY"], width=1000000))
            istack.mylog("ACTIVE: %s\n" % pformat(deps["ACTIVE"], width=1000000))

            # is update stuck ?
            max_retry = istack.cfg.max_retry_ecs_service_running_count
            if max_retry > 0 and failedTasks >= max_retry:
                istack.last_event_timestamp = event.timestamp
                raise IboxErrorECSService(
                    "ECS Service did not stabilize "
                    f"[{failedTasks} >= {max_retry}] - "
                    "cancelling update [ROLLBACK]"
                )

        time.sleep(5)


# get timestamp from last event available
def get_last_timestamp(istack):
    last_event = list(istack.stack.events.all().limit(1))[0]

    return last_event.timestamp


# show all events after specific timestamp and return last event timestamp
def show(istack, timestamp, timedelta="0"):
    event_iterator = istack.stack.events.all()
    event_list = []
    for event in event_iterator:
        if event.timestamp > timestamp:
            event_list.insert(0, event)
        else:
            break
    for event in event_list:
        logtime = timegm(event.timestamp.timetuple())
        istack.mylog(
            event.logical_resource_id
            + " "
            + event.resource_status
            + " "
            + str(datetime.fromtimestamp(logtime))
            + " "
            + str(event.resource_status_reason)
        )
        if (
            event.resource_type == "AWS::ECS::Service"
            and event.resource_status == "UPDATE_IN_PROGRESS"
            and event.resource_status_reason is None
            and istack.stack.stack_status not in istack.cfg.STACK_COMPLETE_STATUS
        ):
            _show_service_update(istack, event, timedelta)

    if len(event_list) > 0:
        return event_list.pop().timestamp

    return timestamp



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="iboxstacksops",
    version="0.5.0",
    author="Mello",
    author_email="mello+python@ankot.org",
    description="AWS Infrastructure in a Box - Stacks management program",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mello7tre/AwsIBoxStackOps",
    packages=["iboxstacksops",],
    package_data={},
    install_requires=["boto3", "prettytable", "PyYAML>=5,==5.*",],
    extras_require={"extra": ["slackclient"],},
    python_requires=">=3.7",
    scripts=["scripts/ibox_stacksops.py",],
    license="OSI Approved :: Open Software License 3.0 (OSL-3.0)",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Open Software License 3.0 (OSL-3.0)",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



#!/usr/bin/env python3
from prettytable import (
    PrettyTable,
    ALL as ptALL,
    FRAME as ptFRAME,
    HEADER as ptHEADER,
    NONE as ptNONE,
)
from . import cfg
from .log import logger
from .common import *


def get(data):
    fields = cfg.fields
    table = PrettyTable()
    table.padding_width = 1
    table.field_names = fields
    for n in data:
        parameter_not_empty = n.get("parameter_not_empty", [])
        table.add_row(
            [
                "Null"
                if (i not in n or not n[i])
                else f"{n[i]} *"
                if i in parameter_not_empty
                else n[i]
                for i in fields
            ]
        )

    table.sortby = fields[0]
    table.reversesort = True
    table.align = "l"

    if cfg.output == "html":
        table.format = True

        return table.get_html_string(fields=fields)

    if cfg.output == "bare":
        table.header = False
        table.border = False
        table.left_padding_width = 0
        table.padding_width = 1

    return table.get_string(fields=fields)



import os
import sys
import yaml
import json
import time
import botocore
from datetime import datetime, timedelta, tzinfo
from pprint import pprint, pformat


class IboxError(Exception):
    pass


class IboxErrorECSService(Exception):
    pass


CLF_YAML_FUNC = (
    "!Ref",
    "!GetAtt",
    "!GetAZs",
)


def yaml_exclamation_mark(dumper, data):
    if data.startswith(CLF_YAML_FUNC):
        tag = data.split(" ")[0]
        value = dumper.represent_scalar(tag, data.replace(f"{tag} ", ""))
    else:
        value = dumper.represent_scalar("tag:yaml.org,2002:str", data)

    return value


yaml.add_representer(str, yaml_exclamation_mark)
